/*
***********************************************************************
NOTES
***********************************************************************
*/
/*

Surface energy for various materials

                                          Surface
Potential         Structure Orientation   Energy        Source
                                          (erg/cm^2)
-------------     --------- -----------   ----------    -------
NiAl (Voter)      B2        100           1842          JR

NiAl (Voter)      B2        100            950          Charlotte's Thesis
NiAl (Voter)      B2        110           1600          Charlotte's Thesis
NiAl (Voter)      B2        111           2000          Charlotte's Thesis
NiAl (Voter)      B2        211           1900          Charlotte's Thesis

RuAl (JR)         B2        100           3203          Charlotte's Thesis
RuAl (JR)         B2        110           2446          Charlotte's Thesis
RuAl (JR)         B2        111           3029          Charlotte's Thesis
RuAl (JR)         B2        210           2998          Charlotte's Thesis
RuAl (JR)         B2        211           2873          Charlotte's Thesis
RuAl (JR)         B2        521           3103          Charlotte's Thesis
*/

/*
***********************************************************************
File Includes
***********************************************************************
*/
using namespace std;
#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <cstring>
#include <complex>
#include "matrixd.h"
#include "strsub.h"
#include "polynom.h"


/*
***********************************************************************
Defines
***********************************************************************
*/
#define NSTR 256
#define N_ECONST 21
#define N_ORIENT  9
#define NDIR 3
#define N_REQ_ARG 6

#define BOOLEAN int
#define TRUE 1
#define FALSE 0

/*  Centimeter - Anstrom conversion  */
#define ANG_FROM_CM 1e+8
#define CM_FROM_ANG 1e-8

/*
Number of roots in polynomial equation
*/
#define NROOT  4

/*  Direction Indices  */
#define X  0
#define Y  1
#define Z  2
#define NDIR 3

#define COMPLEX complex<double>

/*
***********************************************************************
Compile Switches
***********************************************************************
*/
/*  Compile Debug Data Dumps  */
#define DEBUG
#undef  DEBUG

/*  Fix value of stress intensity to mimic results of eltest.cpp */
#define COMPARE_TEST
#undef  COMPARE_TEST

/*
***********************************************************************
Macros
***********************************************************************
*/
#ifdef DEBUG
#define WRITEMATRIX(string,x,n)  \
   printf ("%s\n", string); \
   Write_MatrixNN( (double *) x,n);
#else
#define WRITEMATRIX (string,x,n) ;
#endif

#define SQR(x)  ((q_m=(x))*q_m)
#undef SQR


/*
***********************************************************************
Type Definitions
***********************************************************************
*/
typedef double  Array21_t    [21];
typedef double  Matrix33_t   [3][3];
typedef double  Matrix66_t   [6][6];
typedef double  Matrix3333_t [3][3][3][3];
typedef float   Coord_t;
typedef Coord_t *Coord3_t    [3];

/*
***********************************************************************
Module Variables
***********************************************************************
*/

char CommentPrefix_m[] = "#*";
/*  User Message  */
void Usage(void) {
	printf("%s\n","CRACK elastic.fil orient.fil coord.fil surfe crackx cracky [cratio]");
	printf("%s\n","");
	printf("%s\n","   Calculate particle displacements corresponding to specified Stress");
	printf("%s\n","   Intensity according to elastic crack theory.");
	printf("%s\n","");
	printf("%s\n","elastic.file - file containing elastic constants (erg/cm^3)");
	printf("%s\n","               under 100 010 001 orientation");
	printf("%s\n","orient.fil   - file containing miller indices");
	printf("%s\n","particle.fil - file containing particle types and coordinates (ang)");
	printf("%s\n","surf_e       - surface energy of crack surface (in erg/cm^2)");
	printf("%s\n","crackx       - x coordiante of crack tip (ang)");
	printf("%s\n","cracky       - y coordinate of crack tip (ang)");
	printf("%s\n","cratio       - optional.  Stress Intensity (KI) that program applies to crack");
	printf("%s\n","               is product of CRATIO and the calculated Critcial Stress");
	printf("%s\n","               Intensity (KIc).  Default is 1.  Also affects output format.");
	printf("%s\n","");
	printf("%s\n","OUTPUT FORMAT");
	printf("%s\n","   If CRATIO is specified, then output is list of new particle coordinates");
	printf("%s\n","(in angs).  If CRATIO not specified, output is list of old particles");
	printf("%s\n","coordinates along with displacments needed to obtain particle positions");
	printf("%s\n","when at Critical Stress Intensity.");
	exit(0);
}

/*
Temporary Variable for Calculating SQR()
(See macro SQR above)
*/
#ifdef SQR
static double q_m;
#endif

/*
***********************************************************************
Local Function Prototypes
***********************************************************************
*/
void ReadArrayFromFile
   (char *, double *, int );

void CalculateRotationMatrix
   (double [NDIR][NDIR], double [NDIR][NDIR]);

void Convert_Array21_To_Matrix66
   (Matrix66_t, Array21_t);

void CalculateRotationMatrix
   (Matrix33_t, Matrix33_t);

void ConvertCompliance_Matrix66_To_Matrix3333
   (Matrix3333_t, Matrix66_t);

void RotateMatrix3333
   (Matrix3333_t, Matrix33_t);

void ConvertCompliance_Matrix3333_to_Matrix66
   (Matrix66_t, Matrix3333_t);

void ConvertCompliance_Matrix66_PlaneCompliance_Matrix66
   (Matrix66_t, Matrix66_t);

double PlaneStrainTerm (Matrix66_t S, int i, int j);

void ReadParticle (char *, int *, int **, Coord3_t);

void CalculateDisplacement
   (
   int,
   Coord3_t,
   Coord3_t,
   Matrix66_t,
   double,
   double,
   double
   );

double GetCriticalStressIntensity (double, Matrix66_t);

void Write_MatrixNN (double *Matrix, int n);

BOOLEAN  IsRotationMatrixOK  (Matrix33_t);

/*
***********************************************************************
Main Routine
***********************************************************************
*/

int main (int argc, char *argv[])
	{
	char  ElasticFileName[NSTR];
	char  OrientationFileName[NSTR];
	char  ParticleFileName[NSTR];
	double SurfaceEnergy;
	double AppliedStressIntensity;
	double CriticalStressIntensity;
	double CriticalStressFactor;
   double CrackX;
   double CrackY;
   Array21_t    Stiffness_Array21;
   Matrix66_t   Stiffness_Matrix66;
   Matrix66_t   Compliance_Matrix66;
   Matrix66_t   PlaneCompliance_Matrix66;
   Matrix3333_t Compliance_Matrix3333;
   Matrix33_t   Orientation_Matrix33;
   Matrix33_t   Rotation_Matrix33;
   Coord3_t     ParticleCoordinates;
   Coord3_t     ParticleDisplacements;
   int          *ParticleTypes;
   int          NumParticles;
   int          ParticleIndex;
   BOOLEAN      RotationMatrixOK;
   BOOLEAN      WriteDisplacementFlag;

   /*  Write Usage Message if not enough parameters  */
   if (argc<N_REQ_ARG+1) Usage();


   /*  Copy File Names  */
   strncpy (ElasticFileName,     argv[1], NSTR);
   strncpy (OrientationFileName, argv[2], NSTR);
   strncpy (ParticleFileName,    argv[3], NSTR);

   /*  Save input values  */
   SurfaceEnergy = atof (argv[4]);
   CrackX        = atof (argv[5]);
   CrackY        = atof (argv[6]);
   if (7<argc)
      CriticalStressFactor = atof (argv[7]);
   else
      CriticalStressFactor = 0.0;

   /*  Test Input Values  */
   if (SurfaceEnergy<0)
      {
      printf ("ERROR:  Surface Energy (%le erg/cm^2) is less than 0.\n",
         SurfaceEnergy);
      exit(1);
      }

   if (CriticalStressFactor<0)
      {
      printf ("ERROR:  Critical Stress Factor (%le) is less than 0.\n",
         CriticalStressFactor);
      exit(1);
      }

   /*  Flag to write particle displacements  */
   WriteDisplacementFlag = (CriticalStressFactor<=0);

   /*  Read Elastic Constants  */
   ReadArrayFromFile
      (ElasticFileName, (double *) Stiffness_Array21, N_ECONST);

   /*  Read Orientation  */
   ReadArrayFromFile
      (OrientationFileName, (double *) Orientation_Matrix33, NDIR*NDIR);

   /*  Convert Stiffness Voight Array to Voight Matrix  */
   Convert_Array21_To_Matrix66
      (Stiffness_Matrix66, Stiffness_Array21);

#ifdef DEBUG
   WRITEMATRIX ("Input Stiffness", Stiffness_Matrix66, 6);
#endif

   /*  Convert Stiffness to Compliance  */
   memcpy
      (Compliance_Matrix66, Stiffness_Matrix66, sizeof(Compliance_Matrix66));

   /*  ... invert Compliance Matrix in place  */
   matrixinv ( (double *) Compliance_Matrix66, 6, 6);

#ifdef DEBUG
   /*  Test accuracy of matrix inversion  */
   printf ("Product of Stiffness and Compliance.\n");
   {
   int Row;
   int Col;
   int Index;
   double Element;

   for (Row=0;Row<6;Row++)
      {
      for (Col=0;Col<6;Col++)
         {
         Element = 0.0;
         for (Index=0;Index<6;Index++)
            Element +=
               Compliance_Matrix66[Row][Index]*Stiffness_Matrix66[Index][Col];
         printf (" %7.4lf", Element);
         }
      printf ("\n");
      }
   }
#endif

#ifdef DEBUG
   WRITEMATRIX ("Derived Compliance", Compliance_Matrix66, 6);
#endif

   /*  Calculate Rotation Matrix from Elastic Constants  */
   CalculateRotationMatrix (Rotation_Matrix33, Orientation_Matrix33);

   /*
   Check rotational matrix
   - if not then orientation vectors not orthonormal
   */
   RotationMatrixOK = IsRotationMatrixOK  (Rotation_Matrix33);
   if (!RotationMatrixOK)
      {
      int Row;
      int Col;

      printf ("ERROR:  Orientation vectors are not orthonormal.\n");
      /*  Print current orientation matrix  */
      for (Row=0; Row<3; Row++)
         {
         printf ("[");
         for (Col=0; Col<3; Col++)
            printf (" %6.2lf", Orientation_Matrix33[Row][Col]);
         printf ("]\n");
         }
      exit(1);
      }

   /*  Convert Compliance constants from Voight format to "native" format  */
   ConvertCompliance_Matrix66_To_Matrix3333
      (Compliance_Matrix3333, Compliance_Matrix66);

   /*  Rotate Native Compliance Matrix */
   RotateMatrix3333 (Compliance_Matrix3333, Rotation_Matrix33);

   /*  Convert Native Compliance Matrix back to Voight Matrix  */
   ConvertCompliance_Matrix3333_to_Matrix66
      (Compliance_Matrix66, Compliance_Matrix3333);

#ifdef DEBUG
   WRITEMATRIX ("Rotated Compliance", Compliance_Matrix66, 6);
#endif

   /*  Get Plane Strain Compliances  */
   ConvertCompliance_Matrix66_PlaneCompliance_Matrix66
      (PlaneCompliance_Matrix66, Compliance_Matrix66);

#ifdef DEBUG
   WRITEMATRIX ("Plane Strain Compliance", PlaneCompliance_Matrix66, 6);
#endif

   /*  Read in coordinates and Calculate Displacemenets  */
	ReadParticle
      (ParticleFileName, &NumParticles, &ParticleTypes, ParticleCoordinates);

   /*  Allocate Particle Displacements  */
   ParticleDisplacements[X] =
      (Coord_t *) calloc (NumParticles, sizeof(Coord_t));
   ParticleDisplacements[Y] =
      (Coord_t *) calloc (NumParticles, sizeof(Coord_t) );
   ParticleDisplacements[Z] =
      (Coord_t *) calloc (NumParticles, sizeof(Coord_t) );
   if (ParticleDisplacements[Z]==NULL)
      {
      printf ("ERROR:  Not enough memory to allocate particles displacements.\n");
      return 1;
      }

   /*  Calculate Critical Stress Intensity  */
   CriticalStressIntensity =
      GetCriticalStressIntensity
         (SurfaceEnergy, PlaneCompliance_Matrix66);

	/*  Set Applied Stress Intensity  */
	if (CriticalStressFactor>0)
		AppliedStressIntensity = CriticalStressFactor * CriticalStressIntensity;
	else
		AppliedStressIntensity = CriticalStressIntensity;

	/*  Calculate particle displacements  */
	CalculateDisplacement
		(
		NumParticles, ParticleCoordinates, ParticleDisplacements,
		PlaneCompliance_Matrix66, CrackX, CrackY,
		AppliedStressIntensity
      );

	/*  Write Stress Intensity as comments  */
	printf ("# Theoretical Critical Stress Intensity (erg/cm^2.5)%13.6e\n", 
		CriticalStressIntensity);
   printf ("# Applied Stress Intensity              (erg/cm^2.5)%13.6e\n",
		AppliedStressIntensity);

   /*  Write displacements  */
   if (WriteDisplacementFlag)
      {
      printf ("DISP %i\n", NumParticles);
      for (ParticleIndex=0; ParticleIndex<NumParticles; ParticleIndex++)
         {
         printf
            (
            "%4i %10.4f %10.4f %10.4f %10.4f %10.4f %10.4f\n",
            ParticleTypes           [ParticleIndex],
            ParticleCoordinates  [X][ParticleIndex],
            ParticleCoordinates  [Y][ParticleIndex],
            ParticleCoordinates  [Z][ParticleIndex],
            ParticleDisplacements[X][ParticleIndex],
            ParticleDisplacements[Y][ParticleIndex],
            ParticleDisplacements[Z][ParticleIndex]
            );
         }
      }

   /* .. Write coordinates  */
   else
      {
      printf ("PARTICLE %i\n", NumParticles);
      for (ParticleIndex=0; ParticleIndex<NumParticles; ParticleIndex++)
         {
         printf
            (
            "%4i %10.4f %10.4f %10.4f\n",
            ParticleTypes           [ParticleIndex],
            ParticleCoordinates  [X][ParticleIndex] +
            ParticleDisplacements[X][ParticleIndex],
            ParticleCoordinates  [Y][ParticleIndex] +
            ParticleDisplacements[Y][ParticleIndex],
            ParticleCoordinates  [Z][ParticleIndex] +
            ParticleDisplacements[Z][ParticleIndex]
            );
         }
      }


   /* end program  */
   return 0;
   }


/*
***********************************************************************
Local Functions
***********************************************************************
*/
void ReadArrayFromFile
(
char *InputFileName,
double *Array,
int    NumElements
)
   {
   FILE *InputFile;
   char InputString[NSTR];
   char *TempString;
   int  NumberRead;

   InputFile = fopen (InputFileName, "rt");

   if (InputFile==NULL)
      {
      printf ("ERROR:  Cannot open data file %s.\n",
         InputFileName);
      exit(1);
      }

   NumberRead = 0;
   while (NumberRead < NumElements)
      {
      /*  Read one line from file  */
      fgets (InputString, NSTR, InputFile);

      /*  Skip comment lines  */
      if (! IsComment (InputString, CommentPrefix_m) )
         {
         TempString = InputString;
         SkipWhiteSpace (&TempString);
         while (TempString[0] != '\0')
            {
            /*  Test for too many numbers  */
            if (NumberRead >= NumElements)
               {
               printf ("ERROR:  Too many numbers in data file %s.\n",
                  InputFileName);
               exit(2);
               }
            Array[NumberRead] = dblstr (&TempString);
            NumberRead++;
            }
         }
      }
   /*  Close input file  */
   fclose (InputFile);
   }

/*
Read particles from file and allocate array space.
File format:
   First line is number of particles.
   Succeeding lines are coordinates
*/
void ReadParticle
(
char *ParticleFileName,
int *NumParticles,
int **ParticleTypes,
Coord3_t ParticleCoordinates
)
   {
   FILE *InputFile;
   char InputString[NSTR];
   char *TempString;
   char *FileStatus;
   int  ParticleIndex;

   InputFile = fopen (ParticleFileName, "rt");

   if (InputFile==NULL)
      {
      printf ("ERROR:  Cannot open input file %s.\n", ParticleFileName);
      exit(1);
      }


   /*  Get first non-comment line  */
   FileStatus = fgets (InputString, NSTR, InputFile);
   while (FileStatus!=NULL && IsComment (InputString, CommentPrefix_m))
      FileStatus = fgets (InputString, NSTR, InputFile);

   if (FileStatus==NULL)
      {
      printf ("ERROR:  Premature end-of-file in particle file %s.\n",
         ParticleFileName);
      exit(1);
      }


   /*  Break string into tokens until non-zero value numeric value found  */
   *NumParticles = 0;
   TempString = InputString;
   while (TempString[0] != '\0' && *NumParticles==0)
      {
      *NumParticles = intstr (&TempString);
      }

   /*  Test for number of particles zero  */
   if (*NumParticles==0)
      {
      printf ("ERROR: Number of particles in coordinate file %s is zero.\n",
         ParticleFileName);
      exit(1);
      }

   /*  Allocate storage for particle coordinates  */
   ParticleCoordinates[X] =
      (Coord_t *) calloc ((*NumParticles), sizeof(Coord_t));
   ParticleCoordinates[Y] =
      (Coord_t *) calloc ((*NumParticles), sizeof(Coord_t));
   ParticleCoordinates[Z] =
      (Coord_t *) calloc ((*NumParticles), sizeof(Coord_t));
   *ParticleTypes =
      (int *) calloc ( (*NumParticles), sizeof(int) );
   if (*ParticleCoordinates == NULL || *ParticleTypes == NULL )
      {
      printf ("ERROR:  Not enough memory to store %i coordinates.\n",
         *NumParticles);
      exit(1);
      }

   /*  Initialize particle index  */
   ParticleIndex = 0;

   /*  Read non-comment lines of particle file  */
   do
      {
      /*  Read next line  */
      FileStatus = fgets (InputString, NSTR, InputFile);
      /*  Read coordinates from line if present and not a comment  */
      if (FileStatus!=NULL && !IsComment (InputString, CommentPrefix_m))
         {
         /*  Parse string into 3 numeric tokens  */
         TempString = InputString;
         (*ParticleTypes    )    [ParticleIndex] = intstr (&TempString);
         ParticleCoordinates  [X][ParticleIndex] = dblstr (&TempString);
         ParticleCoordinates  [Y][ParticleIndex] = dblstr (&TempString);
         ParticleCoordinates  [Z][ParticleIndex] = dblstr (&TempString);
         ParticleIndex++;
         }
      } while (FileStatus!=NULL  && ParticleIndex<*NumParticles);

   /*  Close input file  */
   fclose (InputFile);

   /*  Test for all particles present  */
   if (ParticleIndex<*NumParticles)
      {
      printf ("ERROR:  Premature end of particle file %s.\n",
         ParticleFileName);
      exit (1);
      }

   }


/*  Convert Voight Array to Voight Matrix  */
void Convert_Array21_To_Matrix66
(
Matrix66_t Matrix66,
Array21_t  Array21
)
   {
   int i,j,k;

   for (i=0;i<6;i++)
   for (j=0;j<6;j++)
      {
      if (i>j)
         k = i*(i+1)/2 + j;
      else
         k = j*(j+1)/2 + i;
      Matrix66[i][j] = Array21[k];
      }
   }

/*  Calculate Rotation Matrix from Orientation Matrix  */
void CalculateRotationMatrix
   (Matrix33_t RotationMatrix, Matrix33_t Orientation)
   {
   int Row;
   int Col;
   double q;
   double Magnitude;
   double InvMagnitude;

   /*  Normalize rows  */
   for (Row=0; Row<3; Row++)
      {

      /*  Calculate magnitude of Orientation row  */
      Magnitude = (q=Orientation[Row][0])*q;
      for (Col=1; Col<3; Col++)
         {
         Magnitude += (q=Orientation[Row][Col])*q;
         }

      /*  Test for Magnitude zero  */
      if (Magnitude==0)
         {
         printf ("ERROR:  Row %i of orientation matrix is all zero.\n", Row);
         exit(3);
         }

      /*  Normalize and place in Rotation Matrix */
      InvMagnitude = 1.0/sqrt(Magnitude);
      for (Col=0; Col<3; Col++)
         {
         RotationMatrix[Row][Col] = Orientation[Row][Col] * InvMagnitude;
         }
      }
   }

/* CONVERT index pair to Voight index  */
int voight (int i, int j) {
   if (i==j)
      return (i);
   else
      return(6-i-j);
}


/*  Convert Voight Elastic Constants to Native format  */
void ConvertCompliance_Matrix66_To_Matrix3333
(
Matrix3333_t Matrix3333,
Matrix66_t   Matrix66
)
   {
   int i,j;
   int i0,j0,k0,l0;
   double Element;

   /*  FORM 4'th RANK TENSOR  */
   for (i0=0;i0<3;i0++)
      {
      for (j0=0;j0<3;j0++)
      for (k0=0;k0<3;k0++)
      for (l0=0;l0<3;l0++)
         {

         /*  Get Voight indices for elastic constant  */
         i = voight(i0,j0);
         j = voight(k0,l0);

         /*  Copy current element from Voight Array to Native Matrix  */
         Element = Matrix66[i][j];

         /*
         Note:  If doing compliance rather than stiffness,
                multiply Native[i0][j0][k0][l0] once by 0.5 if i>=3,
                and once by 0.5if j>=3 (twice if both conditions true).
         */
         if (i>=3)
            {
            if (j>=3)
               Element *= 0.25;
            else
               Element *= 0.5;
            }
         else if (j>=3)
            {
            Element *= 0.5;
            }


         Matrix3333[i0][j0][k0][l0] = Element;
         }
      }
   }


void RotateMatrix3333
(
Matrix3333_t Matrix3333,
Matrix33_t   Rotation_Matrix33
)
   {
   double NewElement;
   Matrix3333_t TempMatrix3333;
   int i0,j0,k0,l0;
   int i1,j1,k1,l1;

   /*  Copy NativeMatrix  */
   memcpy (TempMatrix3333, Matrix3333, sizeof(TempMatrix3333) );

   /*  ROTATE 4'TH RANK TENSOR  */
   for (i0=0;i0<3;i0++)
   for (j0=0;j0<3;j0++)
   for (k0=0;k0<3;k0++)
   for (l0=0;l0<3;l0++)
      {
      NewElement = 0.0;
      for (i1=0;i1<3;i1++)
      for (j1=0;j1<3;j1++)
      for (k1=0;k1<3;k1++)
      for (l1=0;l1<3;l1++)
         NewElement +=
            TempMatrix3333[i1][j1][k1][l1]
            * Rotation_Matrix33[i0][i1]
            * Rotation_Matrix33[j0][j1]
            * Rotation_Matrix33[k0][k1]
            * Rotation_Matrix33[l0][l1];
      Matrix3333[i0][j0][k0][l0] = NewElement;
      }
   }

void ConvertCompliance_Matrix3333_to_Matrix66
(
Matrix66_t Matrix66,
Matrix3333_t Matrix3333
)
   {
   Matrix66_t ElementCount;
   int i,j;
   int i0,j0,k0,l0;

   /*  Initialize Matrix 66 and Count of Elements added */
   for (i=0;i<6;i++)
   for (j=0;j<6;j++)
      {
      Matrix66[i][j] = 0.0;
      ElementCount[i][j] = 0.0;
      }

   for (i0=0;i0<3;i0++)
   for (j0=0;j0<3;j0++)
   for (k0=0;k0<3;k0++)
   for (l0=0;l0<3;l0++)
      {
      i = voight(i0,j0);
      j = voight(k0,l0);
      Matrix66[i][j] += Matrix3333[i0][j0][k0][l0];
      ElementCount[i][j]++;
      }

   /*  Average each element by number of "visits"  */
   for (i=0;i<6;i++)
   for (j=0;j<6;j++)
      {
      /*  Test that every element in matrix66 has been visited a least once  */
      if (ElementCount[i][j]==0)
         {
         printf ("INTERNAL ERROR:  count of %i %i is zero\n",i,j);
         exit(1);
         }

      Matrix66[i][j] /= ElementCount[i][j];

      /*  Conversion factors for Voight Definition  */
      if (i>2)
         {
         if (j>2)
            {
            Matrix66[i][j] *= 4;
            }
         else
            {
            Matrix66[i][j] *= 2;
            }
         }
      else if (j>2)
         Matrix66[i][j] *= 2;
      }

   }

void ConvertCompliance_Matrix66_PlaneCompliance_Matrix66
(
Matrix66_t PS,
Matrix66_t S
)
   {
   /*
   Test for inverse of S[3][3]
   (Needed in PlaneStrainTerm calculation)
   */
   if (S[2][2]==0.0)
      {
      printf ("INTERNAL ERROR:  S[2][2] is zero.\n");
      exit(4);
      }

   /*  Set PS equal to S  */
   memcpy (PS, S, sizeof(Matrix66_t) );

   /*  Reset individual terms of PS  */
   PS[0][0] = PlaneStrainTerm (S, 0, 0);
   PS[1][1] = PlaneStrainTerm (S, 1, 1);
   PS[5][5] = PlaneStrainTerm (S, 5, 5);
   PS[0][1] =
   PS[1][0] = PlaneStrainTerm (S, 0, 1);
   PS[1][5] =
   PS[5][1] = PlaneStrainTerm (S, 1, 5);
   PS[5][0] =
   PS[0][5] = PlaneStrainTerm (S, 5, 0);
   }

double PlaneStrainTerm (Matrix66_t S, int i, int j)
   {
   return  S[i][j] - S[i][2] * S[j][2] / S[2][2];
   }


void Write_MatrixNN (double *Matrix, int n)
   {
   int i,j;

   for (i=0; i<n; i++)
      {
      for (j=0; j<n; j++)
         printf ("%12.4le ", Matrix[i*n + j]);
      printf ("\n");
      }
   }



/*
Calculate displacement of particles due to perfectly elastic crack
under applied stress
PARAMETERS
 NumParticles                 Number of particles in ParticleCoordinates
 ParticleCoordinates          Particle Coordinates      (ang)
 ParticleDisplacements        Particle Displacements    (ang)
 PlaneCompliance_Matrix66     Plane Strain Compliance   (erg/cm^3)
 CrackX                       X coordinate of crack tip (ang)
 CrackY                       Y coordinate of crack tip (ang)
 StressIntensity              Stress Intensity Factor   (erg/cm^2.5)
*/
void CalculateDisplacement
(
int         NumParticles,
Coord3_t    ParticleCoordinates,
Coord3_t    ParticleDisplacements,
Matrix66_t  PlaneCompliance_Matrix66,
double      CrackX,
double      CrackY,
double      StressIntensity
)
   {
   double       (*S)[6];
   double       sReal[NROOT];
   double       sImag[NROOT];
   double       *Coefficients;
   double       dX;
   double       dY;
   double       Radius;
   double       Angle;
   double       Cosine;
   double       Sine;
   double       Factor2;
   int          RootIndex;
   int          ParticleIndex;
   COMPLEX      Factor1;
   COMPLEX      sComplex[NROOT];
   COMPLEX      p1;
   COMPLEX      p2;
   COMPLEX      q1;
   COMPLEX      q2;
   COMPLEX      s1;
   COMPLEX      s2;


   /*  Allocate space for Coefficients  */
   Coefficients = (double *) calloc (NROOT+1, sizeof(double));
   if (Coefficients==NULL)
      {
      printf ("ERROR:  No memory space for Coefficients array.\n");
      exit(1);
      }

   /*  Calculate values for s  (Solutions to Polynomial Equation)  */
   Coefficients[0] =          PlaneCompliance_Matrix66[1][1];
   Coefficients[1] =  -2.0  * PlaneCompliance_Matrix66[1][5];
   Coefficients[2] = (
                       2.0  * PlaneCompliance_Matrix66[0][1]
                            + PlaneCompliance_Matrix66[5][5]
                      );
   Coefficients[3] =         -PlaneCompliance_Matrix66[0][5];
   Coefficients[4] =          PlaneCompliance_Matrix66[0][0];

   SolvePolynomial (Coefficients, NROOT, sReal, sImag);

   /*  Convert from separate real and imag components to complex type  */
   for (RootIndex=0; RootIndex<NROOT; RootIndex++)
      {
      sComplex[RootIndex] = COMPLEX (sReal[RootIndex], sImag[RootIndex] );
      }

#ifdef DEBUG
   {
   double mag;
   for (RootIndex=0; RootIndex<4; RootIndex++)
      {
      mag = sqrt (pow(sReal[RootIndex],2) + pow(sImag[RootIndex],2));
      printf ("%i   (%12.4le,%12.4lei)  %12.4le\n",
         RootIndex, sReal[RootIndex], sImag[RootIndex], mag);
      }
   }
#endif

   /*
   Calculate variables p1, p2, q1, q2 for displacmenet equation
   (S is pointer to double[6])
   */
   s1  = sComplex[0];
   s2  = sComplex[2];
   Factor1 = 1.0 / (s1 - s2);

   S   = PlaneCompliance_Matrix66;
   p1  = S[0][0]*s1*s1 + S[0][1] - S[0][5] * s1;
   p2  = S[0][0]*s2*s2 + S[0][1] - S[0][5] * s2;
   q1  = S[0][1]*s1*s1 + S[1][1] - S[1][5] * s1;
   q2  = S[0][1]*s2*s2 + S[1][1] - S[1][5] * s2;
   q1 /= s1;
   q2 /= s2;


   /*  Test value of StressIntensity to give results to match eltest.cpp  */
#ifdef COMPARE_TEST
   StressIntensity = 1000.0 * 24076.42;
#endif

   /*  Calculate displacement for all particles  */
   for (ParticleIndex=0; ParticleIndex<NumParticles; ParticleIndex++)
      {
      dX = ParticleCoordinates[X][ParticleIndex] - CrackX;
      dY = ParticleCoordinates[Y][ParticleIndex] - CrackY;
      Radius = sqrt  (dX*dX + dY*dY);

      /*  Test for particle on crack tip  */
      if (Radius==0)
         {
         ParticleDisplacements[X][ParticleIndex] = 0.0;
         ParticleDisplacements[Y][ParticleIndex] = 0.0;
         }


      /*  Particle ok  */
      else
         {
         /*  Initialize variables  */
         Angle  = atan2 (dY, dX);
         Cosine = cos(Angle);
         Sine   = sin(Angle);

         /*  Convert Radius to cm from ang  */
         Radius *= CM_FROM_ANG;

         Factor2 = StressIntensity * sqrt (2 * Radius);

         /*  Calculate X and Y displacement  */
         ParticleDisplacements[X][ParticleIndex] =
            real
               (
               Factor1 *
                  (
                  s1 * p2 * sqrt ( COMPLEX ( Cosine + s2*Sine) ) -
                  s2 * p1 * sqrt ( COMPLEX ( Cosine + s1*Sine) )
                  )
               )
               * Factor2;

         ParticleDisplacements[Y][ParticleIndex] =
            real
               (
               Factor1 *
                  (
                  s1 * q2 * sqrt ( COMPLEX ( Cosine + s2*Sine) ) -
                  s2 * q1 * sqrt ( COMPLEX ( Cosine + s1*Sine) )
                  )
               )
               * Factor2;
         }


      /*  Convert to ang from cm  */
      ParticleDisplacements[X][ParticleIndex] *= ANG_FROM_CM;
      ParticleDisplacements[Y][ParticleIndex] *= ANG_FROM_CM;

      /*  Set Z displacement to 0  */
      ParticleDisplacements[Z][ParticleIndex] = 0.0;
      }

   /*  Release storage  */
   free (Coefficients);

   }

/*
Calculate critical stress intensity for perfectly brittle material under
Mode I loading, given surface energy and plane strain compliances.
UNITS:
 SurfaceEnergy erg/cm^2
 Compliance    erg/cm^3
*/
double GetCriticalStressIntensity (double SurfaceEnergy, Matrix66_t Compliance)
   {
   double Term1, Term2, Term3, Term4;

   /*  Test for Compliance[0][0] equal to zero  */
   if (Compliance[0][0]==0)
      {
      printf ("ERROR (GetCriticalStressIntensity):  Compliance term S11 is zero.\n");
      exit(1);
      }

   Term1 = 2 * SurfaceEnergy / M_PI;
   Term2 = Compliance[0][0] * Compliance[1][1] / 2.0;
   Term3 = Compliance[1][1] / Compliance[0][0];
   Term4 =  ( 2 * Compliance[0][1] + Compliance[5][5] ) /
            ( 2 * Compliance[0][0]);

   return sqrt(Term1) * pow (Term2 * (sqrt(Term3) + Term4), -0.25);
   }

BOOLEAN IsRotationMatrixOK (Matrix33_t Rotation)
   {
   double Element;
   double Error;
   int    Row;
   int    Col;
   int    Index;
   double MaxAllowedError = 1e-5;
   Matrix33_t IndentityMatrix =
      {
      {1, 0, 0},
      {0, 1, 0},
      {0, 0, 1}
      };

   /*  Loop over elements of product matrix (R ~R)  */
   for (Row=0;Row<3;Row++)
   for (Col=0;Col<3;Col++)
      {
      Element = 0.0;
      for (Index=0;Index<3;Index++)
         {
         Element += Rotation[Row][Index]*Rotation[Col][Index];
         }
      Error = Element - IndentityMatrix[Row][Col];
      if (fabs (Error) > MaxAllowedError)
         return FALSE;
      }

   /*  If gotten this far then matrix OK  */
   return TRUE;

   }
